@@ -3,6 +3,7 @@
 #include "lib.h"
 #include "lib-signals.h"
 #include "ioloop.h"
+#include "array.h"
 #include "hostpid.h"
 #include "home-expand.h"
 #include "env-util.h"
@@ -49,12 +50,16 @@ void (*hook_mail_storage_created)(struct mail_storage *storage) = NULL;
 
 /* FIXME: these two should be in some context struct instead of as globals.. */
 static const char *default_mailbox_name = NULL;
+static bool saved_mail = FALSE;
 static bool tried_default_save = FALSE;
 static bool no_mailbox_autocreate = FALSE;
 
 static struct module *modules;
 static struct ioloop *ioloop;
 
+static pool_t plugin_pool;
+static array_t ARRAY_DEFINE(plugin_envs, const char *);
+
 static void sig_die(int signo, void *context __attr_unused__)
 {
 	/* warn about being killed because of some signal, except SIGINT (^C)
@@ -142,9 +147,10 @@ int deliver_save(struct mail_storage *storage, const char *mailbox,
 	msgid = msgid == NULL ? "" : str_sanitize(msgid, 80);
 	mailbox_name = str_sanitize(mailbox_get_name(box), 80);
 
-	if (ret == 0)
+	if (ret == 0) {
+		saved_mail = TRUE;
 		i_info("msgid=%s: saved mail to %s", msgid, mailbox_name);
-	else {
+	} else {
 		const char *error;
 		bool syntax, temp;
 
@@ -205,9 +211,13 @@ static void config_file_init(const char *path)
 	struct istream *input;
 	const char *key, *value;
 	char *line, *p, quote;
-	int fd, sections = 0, lda_section = FALSE, pop3_section = FALSE;
+	int fd, sections = 0;
+	bool lda_section = FALSE, pop3_section = FALSE, plugin_section = FALSE;
 	size_t len;
 
+	plugin_pool = pool_alloconly_create("Plugin strings", 512);
+	ARRAY_CREATE(&plugin_envs, default_pool, const char *, 16);
+
 	fd = open(path, O_RDONLY);
 	if (fd < 0)
 		i_fatal_status(EX_CONFIG, "open(%s) failed: %m", path);
@@ -250,16 +260,19 @@ static void config_file_init(const char *path)
 		value = p = strchr(line, '=');
 		if (value == NULL) {
 			if (strchr(line, '{') != NULL) {
-				if (strcmp(line, "protocol lda {") == 0 ||
-				    strcmp(line, "plugin {") == 0)
+				if (strcmp(line, "protocol lda {") == 0)
 					lda_section = TRUE;
-				if (strcmp(line, "protocol pop3 {") == 0)
+				else if (strcmp(line, "plugin {") == 0) {
+					plugin_section = TRUE;
+					lda_section = TRUE;
+				} else if (strcmp(line, "protocol pop3 {") == 0)
 					pop3_section = TRUE;
 				sections++;
 			}
 			if (*line == '}') {
 				sections--;
 				lda_section = FALSE;
+				plugin_section = FALSE;
 				pop3_section = FALSE;
 			}
 			continue;
@@ -297,7 +310,16 @@ static void config_file_init(const char *path)
 			continue;
 		}
 
-		env_put(t_strconcat(t_str_ucase(key), "=", value, NULL));
+		if (!plugin_section) {
+			env_put(t_strconcat(t_str_ucase(key), "=",
+					    value, NULL));
+		} else {
+			/* %variables need to be expanded.
+			   store these for later. */
+			value = p_strconcat(plugin_pool,
+					    t_str_ucase(key), "=", value, NULL);
+			array_append(&plugin_envs, &value, 1);
+		}
 	}
 	i_stream_unref(&input);
 	t_pop();
@@ -465,6 +487,44 @@ static void open_logfile(const char *username)
 	i_set_failure_timestamp_format(stamp);
 }
 
+static const char *expand_envs(const char *destination)
+{
+        const struct var_expand_table *table;
+	const char *mail_env, *const *envs;
+	unsigned int i, count;
+	string_t *str;
+
+	str = t_str_new(256);
+	table = get_var_expand_table(destination, getenv("HOME"));
+	envs = array_get(&plugin_envs, &count);
+	for (i = 0; i < count; i++) {
+		str_truncate(str, 0);
+		var_expand(str, envs[i], table);
+		env_put(str_c(str));
+	}
+
+	/* get the table again in case plugin provided the home directory
+	   (yea, kludgy) */
+	table = get_var_expand_table(destination, getenv("HOME"));
+
+	/* MAIL comes from userdb, MAIL_LOCATION from dovecot.conf.
+	   We don't want to expand settings coming from userdb. */
+	mail_env = getenv("MAIL");
+	if (mail_env == NULL)  {
+		mail_env = getenv("MAIL_LOCATION");
+		if (mail_env == NULL)  {
+			/* Keep this for backwards compatibility */
+			mail_env = getenv("DEFAULT_MAIL_ENV");
+		}
+		if (mail_env != NULL) {
+			table = get_var_expand_table(destination,
+						     getenv("HOME"));
+			mail_env = expand_mail_env(mail_env, table);
+		}
+	}
+	return mail_env;
+}
+
 static void print_help(void)
 {
 	printf(
@@ -478,8 +538,7 @@ int main(int argc, char *argv[])
 	const char *envelope_sender = DEFAULT_ENVELOPE_SENDER;
 	const char *mailbox = "INBOX";
 	const char *auth_socket, *env_tz;
-	const char *home, *destination, *user, *mail_env, *value;
-        const struct var_expand_table *table;
+	const char *home, *destination, *user, *value, *mail_env;
         enum mail_storage_flags flags;
         enum mail_storage_lock_method lock_method;
 	struct mail_storage *storage, *mbox_storage;
@@ -646,27 +705,13 @@ int main(int argc, char *argv[])
 	if (deliver_set->sendmail_path == NULL)
 		deliver_set->sendmail_path = DEFAULT_SENDMAIL_PATH;
 
+	mail_env = expand_envs(destination);
+
 	dict_client_register();
         duplicate_init();
         mail_storage_init();
 	mail_storage_register_all();
 
-	/* MAIL comes from userdb, MAIL_LOCATION from dovecot.conf.
-	   We don't want to expand settings coming from userdb. */
-	mail_env = getenv("MAIL");
-	if (mail_env == NULL)  {
-		mail_env = getenv("MAIL_LOCATION");
-		if (mail_env == NULL)  {
-			/* Keep this for backwards compatibility */
-			mail_env = getenv("DEFAULT_MAIL_ENV");
-		}
-		if (mail_env != NULL) {
-			table = get_var_expand_table(destination,
-						     getenv("HOME"));
-			mail_env = expand_mail_env(mail_env, table);
-		}
-	}
-
 	module_dir_init(modules);
 
 	/* FIXME: how should we handle namespaces? */
@@ -699,22 +744,22 @@ int main(int argc, char *argv[])
 		i_fatal("mail_set_seq() failed");
 
 	default_mailbox_name = mailbox;
-	ret = deliver_mail == NULL ? 0 :
-		deliver_mail(storage, mail, destination, mailbox);
+	if (deliver_mail != NULL)
+		(void)deliver_mail(storage, mail, destination, mailbox);
 
-	if (ret == 0 || (ret < 0 && !tried_default_save)) {
+	if (!saved_mail && !tried_default_save) {
 		/* plugins didn't handle this. save into the default mailbox. */
 		i_stream_seek(input, 0);
-		ret = deliver_save(storage, mailbox, mail, 0, NULL);
+		(void)deliver_save(storage, mailbox, mail, 0, NULL);
 	}
-	if (ret < 0 && strcasecmp(mailbox, "INBOX") != 0) {
+	if (!saved_mail && strcasecmp(mailbox, "INBOX") != 0) {
 		/* still didn't work. try once more to save it
 		   to INBOX. */
 		i_stream_seek(input, 0);
-		ret = deliver_save(storage, "INBOX", mail, 0, NULL);
+		(void)deliver_save(storage, "INBOX", mail, 0, NULL);
 	}
 
-	if (ret < 0) {
+	if (!saved_mail) {
 		const char *error, *msgid;
 		bool syntax, temporary_error;
 		int ret;
